# -*- coding: utf-8 -*-
"""
   = = = = = = = = = = = = = = = = = = = = =
   =   ocdlib.Opsiclientd                  =
   = = = = = = = = = = = = = = = = = = = = =
   
   opsiclientd is part of the desktop management solution opsi
   (open pc server integration) http://www.opsi.org
   
   Copyright (C) 2010 uib GmbH
   
   http://www.uib.de/
   
   All rights reserved.
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 as
   published by the Free Software Foundation.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
   
   @copyright:	uib GmbH <info@uib.de>
   @author: Jan Schneider <j.schneider@uib.de>
   @license: GNU General Public License version 2
"""

__version__ = '4.0.23'
__fullversion__ = True

# Imports
import sys, os

# Twisted imports
from twisted.internet import reactor

# OPSI imports
from OPSI.Logger import *
from OPSI.Util import *
from OPSI.Util.Message import *
from OPSI.Types import *
from OPSI import System
from OPSI.Object import *

from ocdlib.Exceptions import *
from ocdlib.Events import *
from ocdlib.ControlPipe import ControlPipeFactory, OpsiclientdRpcPipeInterface
from ocdlib.ControlServer import ControlServer
if (os.name == 'nt'):
	from ocdlib.Windows import *
if (os.name == 'posix'):
	from ocdlib.Posix import *
from ocdlib.Localization import _, setLocaleDir, getLanguage
from ocdlib.Config import Config
try:
	from ocdlibnonfree.CacheService import CacheService
except:
	__fullversion__ = False
	from ocdlib.CacheService import CacheService
try:
	from ocdlibnonfree.EventProcessing import EventProcessingThread
except:
	__fullversion__ = False
	from ocdlib.EventProcessing import EventProcessingThread

logger = Logger()
config = Config()

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# -                                            OPSICLIENTD                                            -
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Opsiclientd(EventListener, threading.Thread):
	def __init__(self):
		moduleName = u' %-30s' % (u'opsiclientd')
		logger.setLogFormat(u'[%l] [%D] [' + moduleName + u'] %M   (%F|%N)', object=self)
		logger.debug(u"Opsiclient initiating")
		
		EventListener.__init__(self)
		threading.Thread.__init__(self)
		
		self._startupTime = time.time()
		self._running = False
		self._eventProcessingThreads = []
		self._eventProcessingThreadsLock = threading.Lock()
		self._blockLogin = True
		self._currentActiveDesktopName = {}
		
		self._isRebootTriggered = False
		self._isShutdownTriggered = False
		
		self._actionProcessorUserName = u''
		self._actionProcessorUserPassword = u''
		
		self._statusApplicationProcess = None
		self._blockLoginNotifierPid = None
		
		self._popupNotificationServer = None
		self._popupNotificationLock = threading.Lock()
		
	
	def setBlockLogin(self, blockLogin):
		self._blockLogin = bool(blockLogin)
		logger.notice(u"Block login now set to '%s'" % self._blockLogin)
		
		if (self._blockLogin):
			if not self._blockLoginNotifierPid and config.get('global', 'block_login_notifier'):
				logger.info(u"Starting block login notifier app")
				sessionId = System.getActiveConsoleSessionId()
				while True:
					try:
						self._blockLoginNotifierPid = System.runCommandInSession(
								command = config.get('global', 'block_login_notifier'),
								sessionId = sessionId,
								desktop = 'winlogon',
								waitForProcessEnding = False)[2]
						break
					except Exception, e:
						logger.error(e)
						if (e[0] == 233) and (sys.getwindowsversion()[0] == 5) and (sessionId != 0):
							# No process is on the other end
							# Problem with pipe \\\\.\\Pipe\\TerminalServer\\SystemExecSrvr\\<sessionid>
							# After logging off from a session other than 0 csrss.exe does not create this pipe or CreateRemoteProcessW is not able to read the pipe.
							logger.info(u"Retrying to run command in session 0")
							sessionId = 0
						else:
							logger.error(u"Failed to start block login notifier app: %s" % forceUnicode(e))
							break
		elif (self._blockLoginNotifierPid):
			try:
				logger.info(u"Terminating block login notifier app (pid %s)" % self._blockLoginNotifierPid)
				System.terminateProcess(processId = self._blockLoginNotifierPid)
			except Exception, e:
				logger.warning(u"Failed to terminate block login notifier app: %s" % forceUnicode(e))
			self._blockLoginNotifierPid = None
		
	def isRunning(self):
		return self._running
	
	def waitForGUI(self, timeout=None):
		if not timeout:
			timeout = None
		class WaitForGUI(EventListener):
			def __init__(self):
				self._guiStarted = threading.Event()
				eventGenerator = EventGeneratorFactory(GUIStartupEventConfig("wait_for_gui"))
				eventGenerator.addEventListener(self)
				eventGenerator.start()
			
			def processEvent(self, event):
				logger.info(u"GUI started")
				self._guiStarted.set()
				
			def wait(self, timeout=None):
				self._guiStarted.wait(timeout)
				if not self._guiStarted.isSet():
					logger.warning(u"Timed out after %d seconds while waiting for GUI" % timeout)
				
		WaitForGUI().wait(timeout)
	
	def createActionProcessorUser(self, recreate = True):
		if not config.get('action_processor', 'create_user'):
			return
		
		runAsUser = config.get('action_processor', 'run_as_user')
		if (runAsUser.lower() == 'system'):
			self._actionProcessorUserName = u''
			self._actionProcessorUserPassword = u''
			return
		
		if (runAsUser.find('\\') != -1):
			logger.warning(u"Ignoring domain part of user to run action processor '%s'" % runAsUser)
			runAsUser = runAsUser.split('\\', -1)
		
		if not recreate and self._actionProcessorUserName and self._actionProcessorUserPassword and System.existsUser(username = runAsUser):
			return
		
		self._actionProcessorUserName = runAsUser
		logger.notice(u"Creating local user '%s'" % runAsUser)
		
		self._actionProcessorUserPassword = u'$!?' + unicode(randomString(16)) + u'!/%'
		logger.addConfidentialString(self._actionProcessorUserPassword)
		
		if System.existsUser(username = runAsUser):
			System.deleteUser(username = runAsUser)
		System.createUser(username = runAsUser, password = self._actionProcessorUserPassword, groups = [ System.getAdminGroupName() ])
	
	def deleteActionProcessorUser(self):
		if not config.get('action_processor', 'delete_user'):
			return
		if not self._actionProcessorUserName:
			return
		logger.notice(u"Deleting local user '%s'" % self._actionProcessorUserName)
		if not System.existsUser(username = self._actionProcessorUserName):
			return
		System.deleteUser(username = self._actionProcessorUserName)
		self._actionProcessorUserName = u''
		self._actionProcessorUserPassword = u''
	
	def run(self):
		self._running = True
		self._stopped = False
		
		config.readConfigFile()
		setLocaleDir(config.get('global', 'locale_dir'))
		
		try:
			if __fullversion__:
				logger.comment(u"Opsiclientd version: %s (full)" % __version__)
			else:
				logger.comment(u"Opsiclientd version: %s" % __version__)
			logger.comment(u"Commandline: %s" % ' '.join(sys.argv))
			logger.comment(u"Working directory: %s" % os.getcwd())
			logger.notice(u"Using host id '%s'" % config.get('global', 'host_id'))
			
			self.setBlockLogin(True)
			
			logger.notice(u"Starting control pipe")
			try:
				self._controlPipe = ControlPipeFactory(OpsiclientdRpcPipeInterface(self))
				self._controlPipe.start()
				logger.notice(u"Control pipe started")
			except Exception, e:
				logger.error(u"Failed to start control pipe: %s" % forceUnicode(e))
				raise
			
			logger.notice(u"Starting control server")
			try:
				self._controlServer = ControlServer(
								opsiclientd        = self,
								httpsPort          = config.get('control_server', 'port'),
								sslServerKeyFile   = config.get('control_server', 'ssl_server_key_file'),
								sslServerCertFile  = config.get('control_server', 'ssl_server_cert_file'),
								staticDir          = config.get('control_server', 'static_dir') )
				self._controlServer.start()
				logger.notice(u"Control server started")
			except Exception, e:
				logger.error(u"Failed to start control server: %s" % forceUnicode(e))
				raise
			
			logger.notice(u"Starting cache service")
			try:
				self._cacheService = CacheService(opsiclientd = self)
				self._cacheService.start()
				logger.notice(u"Cache service started")
			except Exception, e:
				logger.error(u"Failed to start cache service: %s" % forceUnicode(e))
				raise
			
			# Create event generators
			createEventGenerators()
			
			for eventGenerator in getEventGenerators():
				eventGenerator.addEventListener(self)
				eventGenerator.start()
				logger.notice(u"Event generator '%s' started" % eventGenerator)
			
			for eventGenerator in getEventGenerators(generatorClass = DaemonStartupEventGenerator):
				eventGenerator.fireEvent()
			
			if getEventGenerators(generatorClass = GUIStartupEventGenerator):
				# Wait until gui starts up
				logger.notice(u"Waiting for gui startup (timeout: %d seconds)" % config.get('global', 'wait_for_gui_timeout'))
				self.waitForGUI(timeout = config.get('global', 'wait_for_gui_timeout'))
				logger.notice(u"Done waiting for GUI")
				
				# Wait some more seconds for events to fire
				time.sleep(5)
			
			if not self._eventProcessingThreads:
				logger.notice(u"No events processing, unblocking login")
				self.setBlockLogin(False)
			
			while not self._stopped:
				time.sleep(1)
			
			for eventGenerator in getEventGenerators(generatorClass = DaemonShutdownEventGenerator):
				eventGenerator.fireEvent()
			
			logger.notice(u"opsiclientd is going down")
			
			for eventGenerator in getEventGenerators():
				logger.info(u"Stopping event generator %s" % eventGenerator)
				eventGenerator.stop()
				eventGenerator.join(2)
			
			for ept in self._eventProcessingThreads:
				logger.info(u"Waiting for event processing thread %s" % ept)
				ept.join(5)
			
			logger.info(u"Stopping cache service")
			if self._cacheService:
				self._cacheService.stop()
				self._cacheService.join(2)
			
			logger.info(u"Stopping control server")
			if self._controlServer:
				self._controlServer.stop()
				self._controlServer.join(2)
			
			logger.info(u"Stopping control pipe")
			if self._controlPipe:
				self._controlPipe.stop()
				self._controlPipe.join(2)
			
			if reactor and reactor.running:
				logger.info(u"Stopping reactor")
				reactor.stop()
				while reactor.running:
					logger.debug(u"Waiting for reactor to stop")
					time.sleep(1)
			
			logger.info(u"Exiting main thread")
			
		except Exception, e:
			logger.logException(e)
			self.setBlockLogin(False)
		
		self._running = False
	
	def stop(self):
		self._stopped = True
	
	def processEvent(self, event):
		
		logger.notice(u"Processing event %s" % event)
		
		eventProcessingThread = None
		self._eventProcessingThreadsLock.acquire()
		try:
			eventProcessingThread = EventProcessingThread(self, event)
			
			# Always process panic events
			if not isinstance(event, PanicEvent):
				for ept in self._eventProcessingThreads:
					if (event.eventConfig.actionType != 'login') and (ept.event.eventConfig.actionType != 'login'):
						raise Exception(u"Already processing an other (non login) event: %s" % ept.event.eventConfig.getName())
					if (event.eventConfig.actionType == 'login') and (ept.event.eventConfig.actionType == 'login'):
						if (ept.getSessionId() == eventProcessingThread.getSessionId()):
							raise Exception(u"Already processing login event '%s' in session %s" \
										% (ept.event.eventConfig.getName(), eventProcessingThread.getSessionId()))
		
		except Exception, e:
			self._eventProcessingThreadsLock.release()
			raise
		
		self.createActionProcessorUser(recreate = False)
		
		self._eventProcessingThreads.append(eventProcessingThread)
		self._eventProcessingThreadsLock.release()
		
		try:
			eventProcessingThread.start()
			eventProcessingThread.join()
			logger.notice(u"Done processing event '%s'" % event)
		finally:
			self._eventProcessingThreadsLock.acquire()
			self._eventProcessingThreads.remove(eventProcessingThread)
			try:
				if not self._eventProcessingThreads:
					self.deleteActionProcessorUser()
			except Exception, e:
				logger.warning(e)
			self._eventProcessingThreadsLock.release()
		
	def getEventProcessingThread(self, sessionId):
		for ept in self._eventProcessingThreads:
			if (int(ept.getSessionId()) == int(sessionId)):
				return ept
		raise Exception(u"Event processing thread for session %s not found" % sessionId)
		
	def processProductActionRequests(self, event):
		logger.error(u"processProductActionRequests not implemented")
	
	def getCurrentActiveDesktopName(self, sessionId=None):
		if not (config.getDict().has_key('opsiclientd_rpc') and config.getDict()['opsiclientd_rpc'].has_key('command')):
			raise Exception(u"opsiclientd_rpc command not defined")
		
		if sessionId is None:
			sessionId = System.getActiveConsoleSessionId()
		rpc = 'setCurrentActiveDesktopName("%s", System.getActiveDesktopName())' % sessionId
		cmd = '%s "%s"' % (config.get('opsiclientd_rpc', 'command'), rpc)
		
		try:
			System.runCommandInSession(command = cmd, sessionId = sessionId, desktop = u"winlogon", waitForProcessEnding = True, timeoutSeconds = 60)
		except Exception, e:
			logger.error(e)
		
		desktop = self._currentActiveDesktopName.get(sessionId)
		if not desktop:
			logger.warning(u"Failed to get current active desktop name for session %d, using 'default'" % sessionId)
			desktop = 'default'
			self._currentActiveDesktopName[sessionId] = desktop
		logger.debug(u"Returning current active dektop name '%s' for session %s" % (desktop, sessionId))
		return desktop
	
	def systemShutdownInitiated(self):
		if not self.isRebootTriggered() and not self.isShutdownTriggered():
			# This shutdown was triggered by someone else
			# Reset shutdown/reboot requests to avoid reboot/shutdown on next boot
			logger.notice(u"Someone triggered a reboot or a shutdown => clearing reboot request")
			self.clearRebootRequest()
	
	def shutdownMachine(self):
		pass
		
	def rebootMachine(self):
		pass
	
	def isRebootTriggered(self):
		if self._isRebootTriggered:
			return True
		return False
		
	def isShutdownTriggered(self):
		if self._isShutdownTriggered:
			return True
		return False
	
	def clearRebootRequest(self):
		pass
		
	def clearShutdownRequest(self):
		pass
		
	def isRebootRequested(self):
		return False
		
	def isShutdownRequested(self):
		return False
	
	def showPopup(self, message):
		port = config.get('notification_server', 'popup_port')
		if not port:
			raise Exception(u'notification_server.popup_port not defined')
		
		notifierCommand = config.get('opsiclientd_notifier', 'command').replace('%port%', forceUnicode(port))
		if not notifierCommand:
			raise Exception(u'opsiclientd_notifier.command not defined')
		notifierCommand += u" -s notifier\\popup.ini"
		
		self._popupNotificationLock.acquire()
		try:
			self.hidePopup()
			
			popupSubject = MessageSubject('message')
			choiceSubject = ChoiceSubject(id = 'choice')
			popupSubject.setMessage(message)
			
			logger.notice(u"Starting popup message notification server on port %d" % port)
			try:
				self._popupNotificationServer = NotificationServer(
								address  = "127.0.0.1",
								port     = port,
								subjects = [ popupSubject, choiceSubject ] )
				self._popupNotificationServer.start()
			except Exception, e:
				logger.error(u"Failed to start notification server: %s" % forceUnicode(e))
				raise
			
			choiceSubject.setChoices([ _('Close') ])
			choiceSubject.setCallbacks( [ self.popupCloseCallback ] )
			
			sessionIds = System.getActiveSessionIds()
			if not sessionIds:
				sessionIds = [ System.getActiveConsoleSessionId() ]
			for sessionId in sessionIds:
				logger.info(u"Starting popup message notifier app in session %d" % sessionId)
				try:
					System.runCommandInSession(
						command = notifierCommand,
						sessionId = sessionId,
						desktop = self.getCurrentActiveDesktopName(sessionId),
						waitForProcessEnding = False)
				except Exception,e:
					logger.error(u"Failed to start popup message notifier app in session %d: %s" % (sessionId, forceUnicode(e)))
		finally:
			self._popupNotificationLock.release()
		
	def hidePopup(self):
		if self._popupNotificationServer:
			try:
				logger.info(u"Stopping popup message notification server")
				self._popupNotificationServer.stop(stopReactor = False)
			except Exception, e:
				logger.error(u"Failed to stop popup notification server: %s" % e)
		
	def popupCloseCallback(self, choiceSubject):
		self.hidePopup()


